package mixer;

/**
 * Created by Tzlil on 12/30/2015.
 */
import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import com.google.protobuf.Message;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.concrete.GlobalCryptoSetup;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeProver;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.protobuf.ConcreteCrypto;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import org.bouncycastle.jce.spec.ECParameterSpec;
import org.bouncycastle.jce.spec.ECPublicKeySpec;
import org.bouncycastle.math.ec.ECPoint;
import org.junit.Test;
import prover.Prover;
import qilin.primitives.RandomOracle;
import qilin.primitives.concrete.DigestOracle;
import qilin.primitives.concrete.ECElGamal;
import qilin.primitives.concrete.ECGroup;
import qilin.primitives.generic.ElGamal;
import qilin.util.Pair;
import verifier.Verifier;
import verifier.VerifyTable;

import java.math.BigInteger;
import java.security.KeyFactory;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class MixingText {

    ECElGamalEncryption encryptor;
    ECGroup group;
    Random random ;
    private BigInteger sk;
    private ECElGamal.SK key;
    private ConcreteCrypto.ElGamalPublicKey serializedPk;

    public final static String ENCRYPTION_KEY_ALGORITHM = "ECDH";
    /**
     * Serialize an El-Gamal public key into a form acceptable by {@link ECElGamalEncryption}
     * @param pk
     * @return
     */
    public static ConcreteCrypto.ElGamalPublicKey serializePk(ECGroup group, ElGamal.PK<ECPoint> pk) {
        ECPoint pkPoint = pk.getPK();
        ECParameterSpec params = group.getCurveParams();

        ECPublicKeySpec pubKeySpec = new ECPublicKeySpec(pkPoint, params);

        try {
            KeyFactory fact = KeyFactory.getInstance(ENCRYPTION_KEY_ALGORITHM,
                    GlobalCryptoSetup.getBouncyCastleProvider());
            PublicKey javaPk = fact.generatePublic(pubKeySpec);
            ConcreteCrypto.ElGamalPublicKey serializedPk = ConcreteCrypto.ElGamalPublicKey.newBuilder()
                    .setSubjectPublicKeyInfo(ByteString.copyFrom(javaPk.getEncoded())).build();

            return serializedPk;
        } catch (Exception e) {
            throw new RuntimeException("Error converting public key!", e);
        }
    }

    public void mixingTest() throws InvalidProtocolBufferException, InvalidKeySpecException {
        // initialization
        random = new Random();
        group = new ECGroup("secp256k1");
        sk = ECElGamal.generateSecretKey(group, random);
        key = new ECElGamal.SK(group, sk);
        serializedPk = serializePk(group, key);
        encryptor = new ECElGamalEncryption();
        encryptor.init(serializedPk);


        Random randomMixer = new Random();
        Random randomProver = new Random();
        RandomOracle randomOracle = new DigestOracle();

        Mix2ZeroKnowledgeVerifier verifier = new Verifier(encryptor,randomOracle);
        Mix2ZeroKnowledgeProver prover = new Prover(randomProver,encryptor,randomOracle,verifier);
        meerkat.crypto.mixnet.Mixer mixer = new Mixer(randomMixer,prover,encryptor);
        ECGroup group = encryptor.getGroup();


        // generate n
        int logN = 7; //  + random.nextInt(8)
        int layers = 2*logN - 1;
        int n = 1 << logN;

        // generate MixerInput
        List<Crypto.RerandomizableEncryptedMessage> mixerInput = new ArrayList<Crypto.RerandomizableEncryptedMessage>();
        Message message;
        for (int i = 0; i < n ; i++){
            message = Mixing.Plaintext.newBuilder()
                    .setData(ByteString.copyFrom(group.encode(group.sample(random))))
                    .build();
            mixerInput.add(encryptor.encrypt(message, encryptor.generateRandomness(random)));
        }

        Pair<Mixing.ZeroKnowledgeProof[][], Crypto.RerandomizableEncryptedMessage[][]> mixerOutput = mixer.mix(mixerInput);

//        assert (VerifyTable.verifyTable(verifier,n,mixerOutput));
    }
}
